
indexing
	description: "Holds the namespace context for resolution during parsing"
	
class NXML_NAMESPACE_CONTEXT

inherit
	ANY
		redefine
			is_equal, copy
		end
		
creation
	make
	
feature {NONE} -- Creation
	
	make is
			-- Make with empty namespace.
		do
			namespace := shared_empty_namespace
			!! by_name.make_default
		end
		
	shared_empty_namespace: NXML_NAMESPACE is
			-- Shared empty namespace.
		local
			a_string: STRING
		once
			!! a_string.make(0)
			!! Result.make (a_string)
		end
	
feature -- ANY

	is_equal (other: like Current): BOOLEAN is
			-- Equality.
		do
			Result := (other = Current)
				or else (
					namespace.is_equal (other.namespace)
					and
					by_name.is_equal (other.by_name)
					)
		end

	copy (other: like Current) is
			-- Copy
		do
			namespace := other.namespace
			by_name.copy (other.by_name)
		end	
		
feature -- Access

	namespace: NXML_NAMESPACE
			-- Current namespace.
			
	named (a_name: STRING): NXML_NAMESPACE is
			-- Named namespace by name.
		require
			not_void: a_name /= Void
			has: has (a_name)
		do
			Result := by_name.item (a_name)
		end
		
	has (a_name: STRING): BOOLEAN is
			-- Has namespace with this name?
		require
			not_void: a_name /= Void
		do
			Result := by_name.has (a_name)
		end

feature {NXML_ANY} -- Setting

	set_namespace (a_namespace: NXML_NAMESPACE) is
			-- Set default namespace.
		do
			namespace := a_namespace
		end
	
	add_named (a_ns: NXML_NAMESPACE; a_name: STRING) is
			-- Add namespace.
		require
			not_void_ns: a_ns /= Void
			not_empty_name: a_name /= Void
		do
			by_name.force (a_ns, a_name)
		end
		
feature {NXML_NAMESPACE_CONTEXT} -- Implementation	
	
	by_name: DS_HASH_TABLE [NXML_NAMESPACE, STRING]
			-- Namespaces indexed by names.

invariant
	not_void_namespace: namespace /= Void
	not_void_by_name: by_name /= Void
	
end
