%{
indexing
	description: "Nenie XML: Parser"
	project: "http://www.nenie.org/eiffel/"
	implements: "XML 1.0 (Second Edition) - W3C Recommendation 6 October 2000"
	license: "http://www.opensource.org/licenses/mit-license.html"
	revision: "$Id: nxml_parser.y,v 1.16 2001/07/13 20:30:59 farnaud Exp $"

deferred class NXML_PARSER

inherit
	YY_PARSER_SKELETON [ANY]
		rename
			make as make_parser,
			report_error as when_error
		undefine
			when_error
		end

	NXML_EVENT_INTERFACE

	NXML_TOKENS
		export
			{NONE} all
		end
	
	NXML_ERRORS
		export
			{NONE} all
		end
	
	KL_IMPORTED_INPUT_STREAM_ROUTINES
		export
			{NONE} all
		end
			

-- TODO:
-- character classes validation
-- pi: should have space after PI_TARGET?
-- getting error messages from scanner
-- check no external entity or DTD for standalone docs
-- xml declaration in external entities

%}

%type <NXML_PARSER_NAME> namespace_name s_tag_name s_tag e_tag attribute
%type <DS_HASH_SET[NXML_PARSER_NAME]> s_tag_trail
%type <STRING> namespace_name_second nm_token doctype_name
%type <STRING> entity_value entity_value_trail
%type <STRING> att_value att_value_trail att_value_trail_item
%type <STRING> value_reference entity_value_reference
%type <STRING> comment_content comment_content_item pi_content_first pi_content_trail pi_content_item
%type <STRING> cdata_body content_text
%type <NXML_ENTITY_DEF> entity_def pe_def
%type <STRING> external_id public_id

%token <STRING> NAME 
%token <STRING> NMTOKEN
%token <STRING> EQ
%token <STRING> SPACE
%token <STRING> CHARDATA
%token <STRING> COMMENT_START COMMENT_END COMMENT_DASHDASH
%token <STRING> PI_START PI_TARGET PI_END PI_RESERVED
%token <STRING> XMLDECLARATION_START
%token <STRING> XMLDECLARATION_END
%token <STRING> XMLDECLARATION_VERSION
%token <STRING> XMLDECLARATION_STANDALONE_YES
%token <STRING> XMLDECLARATION_STANDALONE_NO
%token <STRING> XMLDECLARATION_ENCODING
%token <STRING> CDATA_START
%token <STRING> CDATA_END
%token <STRING> DOCTYPE_START
%token <STRING> DOCTYPE_END
%token <STRING> DOCTYPE_DECLARATION_START DOCTYPE_DECLARATION_END
%token <STRING> DOCTYPE_ELEMENT_EMPTY DOCTYPE_ELEMENT_ANY
%token <STRING> DOCTYPE_ELEMENT DOCTYPE_ATTLIST DOCTYPE_ENTITY DOCTYPE_NOTATION
%token <STRING> DOCTYPE_GROUP_START DOCTYPE_GROUP_END
%token <STRING> DOCTYPE_GROUP_OR DOCTYPE_GROUP_SEQ
%token <STRING> DOCTYPE_GROUP_ZEROMORE DOCTYPE_GROUP_ANY DOCTYPE_GROUP_ONEMORE
%token <STRING> DOCTYPE_PCDATA
%token <STRING> DOCTYPE_PUBLIC
%token <STRING> DOCTYPE_SYSTEM
%token <STRING> DOCTYPE_REQUIRED DOCTYPE_IMPLIED DOCTYPE_FIXED
%token <STRING> DOCTYPE_ATT_CDATA DOCTYPE_ATT_ID 
%token <STRING> DOCTYPE_ATT_IDREF DOCTYPE_ATT_IDREFS
%token <STRING> DOCTYPE_ATT_ENTITY DOCTYPE_ATT_ENTITIES 
%token <STRING> DOCTYPE_ATT_NMTOKEN DOCTYPE_ATT_NMTOKENS
%token <STRING> DOCTYPE_ATT_NOTATION
%token <STRING> DOCTYPE_PERCENT
%token <STRING> DOCTYPE_PEREFERENCE
%token <STRING> ENTITYVALUE_PEREFERENCE
%token <STRING> DOCTYPE_IGNORE DOCTYPE_INCLUDE
%token <STRING> DOCTYPE_NDATA
%token <STRING> DOCTYPE_CONDITIONAL_START
%token <STRING> DOCTYPE_CONDITIONAL_END
%token <STRING> DOCTYPE_CONDITIONAL_IGNORE
%token <STRING> VALUE_START
%token <STRING> VALUE_END
%token <STRING> TAG_START
%token <STRING> TAG_START_END
%token <STRING> TAG_END_EMPTY
%token <STRING> TAG_END
%token <STRING> TAG_NAME_FIRST
%token <STRING> TAG_NAME_ATOM
%token <STRING> TAG_NAME_COLON
%token <STRING> CONTENT_ENTITY CONTENT_CONDITIONAL_END
%token <STRING> ATTRIBUTE_ENTITY ATTRIBUTE_LT
%token <STRING> ENTITY_INVALID INPUT_INVALID

%%
-- This grammar is heavily based on the standard's EBNF productions.
-- Production names are the same as far as sensible and the productions 
-- are in the same order as in the standard text, whose section headings 
-- are included.

-- 2.1 Well formed XML documents

document: prolog element misc_maybe
	;

-- Convenience productions (out of the standard)

-- Namespace handling: to be fully compatible with XML 1.0 which allows 
-- any number of colons in names while still be ready to support namespaces,
-- namespace names (outside of DTD declarations) are handled as a list 
-- of colon-separated atoms.

namespace_name: TAG_NAME_FIRST
		{
			$$ := new_namespace_name
			$$.force_last ($1)
		}
	| TAG_NAME_COLON
		{
			$$ := new_namespace_name
			$$.force_last (empty_string)
		}
	| namespace_name TAG_NAME_COLON namespace_name_second
		{ 
			$$ := $1
			$$.force_last ($3)
		}
	| namespace_name TAG_NAME_COLON
		{
			$$ := $1
			$$.force_last (empty_string)
		}
	;

namespace_name_second: TAG_NAME_ATOM { $$ := $1 }
	| TAG_NAME_FIRST { $$ := $1 }
	;

nm_token: doctype_name { $$ := $1 }
	| NMTOKEN { $$ := $1 }
	;

-- within the DTD name = keywords + ordinary words
-- this is not handled within the scanner because the position 
-- of a word determines whether it is a keyword or not.

doctype_name: NAME { $$ := $1 }
	| DOCTYPE_ATT_CDATA { $$ := $1 }
	| DOCTYPE_ATT_ID { $$ := $1 }
	| DOCTYPE_ATT_IDREF { $$ := $1 }
	| DOCTYPE_ATT_IDREFS { $$ := $1 }
	| DOCTYPE_ATT_ENTITY { $$ := $1 }
	| DOCTYPE_ATT_ENTITIES { $$ := $1 }
	| DOCTYPE_ATT_NMTOKEN { $$ := $1 }
	| DOCTYPE_ATT_NMTOKENS { $$ := $1 }
	| DOCTYPE_ATT_NOTATION { $$ := $1 }
	| DOCTYPE_ELEMENT_EMPTY { $$ := $1 }
	| DOCTYPE_ELEMENT_ANY { $$ := $1 }
	| DOCTYPE_IGNORE { $$ := $1 }
	| DOCTYPE_INCLUDE { $$ := $1 }
	| DOCTYPE_NDATA { $$ := $1 }
	;

maybe_space: -- Empty
	| req_space
	;
	
req_space: space_item
	| req_space space_item
	;

space_item: SPACE
	| DOCTYPE_PEREFERENCE 
		{ 
			-- really applies only in DTD, but token cannot appear in content
			-- and test above would catch, if unintentionally.
			if not in_external_dtd then 
				force_error	(Error_doctype_peref_only_in_dtd) 
			end 
		}
	;
		
-- 2.3 Common syntactic constructs

entity_value: VALUE_START VALUE_END
		{ $$ := empty_string }
	| VALUE_START entity_value_trail VALUE_END
		{ $$ := $2 }
	;

entity_value_trail: CHARDATA
		{ $$ := $1 }
	| entity_value_reference
		{ $$ := $1 }
	| entity_value_trail CHARDATA
		{ $$ := concatenate ($1, $2) }
	| entity_value_trail entity_value_reference
		{ $$ := concatenate ($1, $2) }
	;

entity_value_reference: ENTITYVALUE_PEREFERENCE
		{ $$ := entity_referenced_in_entity_value ($1) }
-- character_reference now CHARDATA
-- bypassed entity handled by scanner
	;
	
att_value: VALUE_START VALUE_END
		{ $$ := empty_string }
	| VALUE_START att_value_trail VALUE_END
		{ $$ := $2 }
	;

att_value_trail: att_value_trail_item
		{ $$ := $1 }
	| att_value_trail att_value_trail_item
		{ $$ := concatenate ($1, $2) }
	;

att_value_trail_item: CHARDATA { $$ := $1 }
	| value_reference { $$ := $1 }
	| ATTRIBUTE_LT { force_error (Error_lt_not_allowed_attribute_value) }
	;
	
value_reference: ATTRIBUTE_ENTITY { $$ := empty_string } -- really handled by read_token
	;

-- 2.5 Comments

comment: COMMENT_START comment_content COMMENT_END
		{ when_comment ($2) }
	| COMMENT_START COMMENT_END
		{ when_comment (empty_string) }
	;

comment_content: comment_content_item
		{ $$ := $1 }
	| comment_content comment_content_item
		{ $$ := concatenate ($1, $2) }
	;
	
comment_content_item: CHARDATA { $$ := $1 }
	| COMMENT_DASHDASH { force_error (Error_no_dash_dash_in_comment) }
	;

-- 2.6 Processing instructions

pi: PI_START PI_TARGET req_space pi_content_first pi_content_trail PI_END
		{ when_processing_instruction ($2, concatenate ($4,$5)) }
	| PI_START PI_TARGET req_space pi_content_first PI_END
		{ when_processing_instruction ($2, concatenate ($4,$5)) }	
	| PI_START PI_TARGET maybe_space PI_END
		{ when_processing_instruction ($2, empty_string) }
	| PI_RESERVED { force_error (Error_pi_xml_reserved) }
	;

pi_content_trail: pi_content_item
		{ $$ := $1 }
	| pi_content_trail pi_content_item
		{ $$ := concatenate ($1, $2) }
	;

pi_content_item: CHARDATA { $$ := $1 }
	| PI_TARGET	{ $$ := $1 }
	| SPACE	{ $$ := $1 }
	;

pi_content_first: CHARDATA { $$ := $1 }
	| PI_TARGET	{ $$ := $1 }
	;
	
-- 2.7 CDATA section

cd_sect: CDATA_START CDATA_END
		{ when_cdata (empty_string) }
	| CDATA_START cdata_body CDATA_END
		{ when_cdata ($2) }
	;

cdata_body: CHARDATA
		{ $$ := $1 }
	| cdata_body CHARDATA
		{ $$ := concatenate ($1, $2) }
	;

-- 2.8 Prolog and DTD

prolog: xml_decl_misc doctype_decl_misc
	;

xml_decl_misc: misc_maybe
	| xml_decl
	| xml_decl misc_trail
	;
	
doctype_decl_misc: -- Empty
	| doctype_decl
	| doctype_decl misc_trail
	;
	
misc_maybe: 
	| misc_trail
	;

misc_trail: misc
	| misc_trail misc
	;

xml_decl:XMLDECLARATION_START version_info xml_decl_opt XMLDECLARATION_END
	| XMLDECLARATION_START error { force_error (Error_xml_declaration) }
	;

xml_decl_opt: maybe_space
	| req_space sd_decl maybe_space
	| req_space encoding_decl maybe_space
	| req_space encoding_decl req_space sd_decl maybe_space
	;
	
version_info: XMLDECLARATION_VERSION
	;

misc: comment
	| pi
	| SPACE
	;

doctype_decl: doctype_decl_internal
	| doctype_decl_external doctype_decl_dtd
	;

doctype_decl_internal: DOCTYPE_START req_space doctype_name maybe_space doctype_decl_declaration DOCTYPE_END
	;
	
doctype_decl_external: DOCTYPE_START req_space doctype_name req_space external_id maybe_space doctype_decl_declaration DOCTYPE_END
	 	{
			debug ("nxml") io.put_string ("[dtd: in]") end
			in_external_dtd := True
			when_external_dtd ($5) 
		}
	;

-- External DTD
doctype_decl_dtd: DOCTYPE_DECLARATION_START text_decl doctype_decl_dtd_content DOCTYPE_DECLARATION_END
		{
			if not in_external_dtd then
				force_error (Error_misplaced_dtd_declaration)
			end
			debug ("nxml") io.put_string ("[dtd: out]") end
			in_external_dtd := False
		}
	;

doctype_decl_dtd_content: -- Empty
	| doctype_decl_dtd_items
	;

doctype_decl_dtd_items: doctype_decl_dtd_item
	| doctype_decl_dtd_items doctype_decl_dtd_item
	;
	
doctype_decl_dtd_item: markup_decl
	| conditional_sect
	;

-- Internal DTD	
doctype_decl_declaration: -- Empty
	| DOCTYPE_DECLARATION_START DOCTYPE_DECLARATION_END maybe_space
	| DOCTYPE_DECLARATION_START doctype_decl_declaration_content DOCTYPE_DECLARATION_END maybe_space
	;
	
doctype_decl_declaration_content: markup_decl
	| doctype_decl_declaration_content markup_decl
	;

-- DTD body
markup_decl: element_decl
	| attlist_decl
	| entity_decl
	| notation_decl
	| PI
	| comment
	| SPACE -- not in XML1.0:29?
	| DOCTYPE_PEREFERENCE
	;
	
-- 2.9 Stand-alone document declaration

sd_decl: XMLDECLARATION_STANDALONE_YES { stand_alone := True }
	| XMLDECLARATION_STANDALONE_NO { stand_alone := False }
	;

-- 3. Element

element: empty_elem_tag
	| s_tag e_tag 
		{ if not ($1).is_equal ($2) then force_error (Error_end_tag_mismatch) end }
	| s_tag content e_tag
		{ if not ($1).is_equal ($3) then force_error (Error_end_tag_mismatch) end }
	| s_tag content error { force_error (Error_element_end_tag) }
	| s_tag error { force_error (Error_element_content) }
	;


-- 3.1 Element tags

s_tag: TAG_START s_tag_name TAG_END
		{
			$$ := $2
			when_start_tag_end
		}
	| TAG_START s_tag_name req_space s_tag_trail TAG_END
		{
			$$ := $2
			when_start_tag_end
		}
	| TAG_START error { force_error (Error_start_tag) }
	;

empty_elem_tag: TAG_START s_tag_name TAG_END_EMPTY
		{
			when_start_tag_end -- makes empty attributes
			when_end_tag ($2)
		}
	| TAG_START s_tag_name req_space s_tag_trail TAG_END_EMPTY
		{
			when_start_tag_end
			when_end_tag ($2)
		}
	;

s_tag_name: namespace_name
		{
			$$ := $1
			when_start_tag ($1)
		}
	;
	
s_tag_trail: attribute
		{
			$$ := new_name_set
			$$.force_new ($1)
		}
	| s_tag_trail req_space attribute
		{
			$$ := $1
			if ($$).has ($3) then
				force_error (Error_attribute_duplicate)
			else
				$$.force_new ($3)
			end
		}
	;

attribute: namespace_name EQ att_value
		{
			$$ := $1
			when_start_tag_attribute ($1, $3) 
		}
	| namespace_name error 
		{ force_error (Error_attribute) }
	;

e_tag: TAG_START_END namespace_name TAG_END
		{
			$$ := $2
			when_end_tag ($2)
		}
	| TAG_START_END error { force_error (Error_end_tag) }
	;

content: content_item
	| content content_item
	;

content_item: content_text { when_content ($1) }
	| cd_sect
	| element
	| pi
	| comment
	| CONTENT_CONDITIONAL_END { force_error (Error_no_cdata_end_in_content) }
	| entity_in_content
	;

entity_in_content: CONTENT_ENTITY -- no action, switched in read_token
	| XMLDECLARATION_END { force_error (Error_entity_xml_declaration) }
	;
	
content_text: CHARDATA -- includes character_reference
		{ $$ := $1 }
	| SPACE 
		{ $$ := $1 }
	;

-- 3.2 Element type declaration

element_decl: DOCTYPE_ELEMENT req_space doctype_name req_space content_spec DOCTYPE_END
	| DOCTYPE_ELEMENT error { force_error (Error_doctype_element) }
	;

content_spec: DOCTYPE_ELEMENT_EMPTY maybe_space
	| DOCTYPE_ELEMENT_ANY maybe_space
	| mixed maybe_space
	| children
	;

children: choice maybe_space
	| choice repetition
	| seq maybe_space
	| seq repetition
	;

cp: doctype_name_space
	| doctype_name repetition
	| choice maybe_space
	| choice repetition
	| seq maybe_space
	| seq repetition
	;

repetition: DOCTYPE_GROUP_ANY maybe_space
	| DOCTYPE_GROUP_ONEMORE maybe_space
	| DOCTYPE_GROUP_ZEROMORE maybe_space
	;

choice: group_start cp group_or choice_trail group_end
	;

choice_trail: cp
	| choice_trail group_or cp
	;

seq: group_start seq_trail group_end
	;

seq_trail: cp
	| seq_trail group_seq cp
	;

mixed: group_start pc_data group_end
	| group_start pc_data group_end DOCTYPE_GROUP_ANY
	| group_start pc_data group_or mixed_trail group_end DOCTYPE_GROUP_ANY
	;

mixed_trail: doctype_name_space
	| mixed_trail group_or doctype_name_space
	;

-- primitives with space

doctype_name_space: DOCTYPE_NAME maybe_space
	;
	
group_start: DOCTYPE_GROUP_START maybe_space
	;
	
group_or: DOCTYPE_GROUP_OR maybe_space
	;
	
group_seq: DOCTYPE_GROUP_SEQ maybe_space
	;
	
group_end: DOCTYPE_GROUP_END
	;

pc_data: DOCTYPE_PCDATA maybe_space
	;
	
-- 3.3 Attribute list declaration

attlist_decl: DOCTYPE_ATTLIST req_space doctype_name maybe_space DOCTYPE_END
	| DOCTYPE_ATTLIST req_space doctype_name attlist_decl_trail maybe_space DOCTYPE_END
	| DOCTYPE_ATTLIST error { force_error (Error_doctype_attribute) }
	;

attlist_decl_trail: att_def
	|  attlist_decl_trail att_def
	;

att_def: req_space doctype_name req_space att_type req_space default_decl
	| req_space doctype_name error { force_error (Error_doctype_attribute_item) }
	;

att_type: DOCTYPE_ATT_CDATA -- string_type
	| att_tokenized_type
	| enumerated_type
	;
	
att_tokenized_type: DOCTYPE_ATT_ID
	| DOCTYPE_ATT_IDREF
	| DOCTYPE_ATT_IDREFS
	| DOCTYPE_ATT_ENTITY
	| DOCTYPE_ATT_ENTITIES
	| DOCTYPE_ATT_NMTOKEN
	| DOCTYPE_ATT_NMTOKENS
	;	

enumerated_type: notation_type
	| enumeration
	;

notation_type: DOCTYPE_ATT_NOTATION req_space group_start notation_type_trail DOCTYPE_GROUP_END
	;

notation_type_trail: doctype_name_space
	| notation_type_trail group_or doctype_name_space
	;

enumeration: group_start enumeration_trail group_end
	;

enumeration_trail: nm_token maybe_space
	| enumeration_trail group_or nm_token maybe_space
	;

default_decl: DOCTYPE_REQUIRED
	| DOCTYPE_IMPLIED
	| DOCTYPE_FIXED req_space att_value
	| att_value
	;

-- 3.4 Conditional section

conditional_sect: include_sect
	| ignore_sect
	| DOCTYPE_CONDITIONAL_START error { force_error (Error_doctype_conditional_section) }
	;

include_sect: include_header DOCTYPE_CONDITIONAL_END
	| include_header doctype_decl_dtd_items DOCTYPE_CONDITIONAL_END
	;

include_header: DOCTYPE_CONDITIONAL_START maybe_space DOCTYPE_INCLUDE maybe_space DOCTYPE_DECLARATION_START
	;
	
ignore_sect: ignore_header ignore_sect_content DOCTYPE_CONDITIONAL_END
	;
	
ignore_header: DOCTYPE_CONDITIONAL_START maybe_space DOCTYPE_IGNORE maybe_space DOCTYPE_DECLARATION_START
		{ scanner.push_start_condition_dtd_ignore }
	;
	
ignore_sect_content: --Empty
	| ignore_sect_items
	;
	
ignore_sect_items: ignore_sect_item
	| ignore_sect_items ignore_sect_item
	;

ignore_sect_item: CHARDATA
	| DOCTYPE_CONDITIONAL_START ignore_sect_content DOCTYPE_CONDITIONAL_END
	;
	
-- 4.1 Reference

-- reference: ENTITY is used directly depending on context
		
-- 4.2 Entity declarations

entity_decl: ge_decl 
	| pe_decl
	| DOCTYPE_ENTITY error { force_error (Error_doctype_entity) }
	;
	
ge_decl: DOCTYPE_ENTITY req_space doctype_name req_space entity_def DOCTYPE_END
		{ when_entity_declared ($3, $5) }
	;
	
pe_decl: DOCTYPE_ENTITY req_space DOCTYPE_PERCENT req_space doctype_name req_space pe_def DOCTYPE_END
		{ when_pe_entity_declared ($5, $7) }
	;
	
entity_def: entity_value maybe_space
		{ $$ := new_literal_entity ($1) }
	| external_id maybe_space
		{ $$ := new_external_entity ($1) }
	| external_id ndata_decl maybe_space
	;

pe_def: entity_value maybe_space
		{ $$ := new_literal_entity ($1) }
	| external_id maybe_space
		{ $$ := new_external_entity ($1) }
	;

external_id: DOCTYPE_SYSTEM
		{ $$ := $1 }
	;

--external_id: DOCTYPE_SYSTEM req_space system_literal
--		{ $$ := $3 }
--	| DOCTYPE_PUBLIC req_space pub_id_literal req_space system_literal
--		{ $$ := $5 }
--	;

ndata_decl: req_space DOCTYPE_NDATA req_space doctype_name -- $2 is 'NDATA'
	;

-- 4.3 Parsed entities

text_decl: -- Empty
	| XMLDECLARATION_START encoding_decl maybe_space XMLDECLARATION_END
	| XMLDECLARATION_START version_info req_space encoding_decl maybe_space XMLDECLARATION_END
	| XMLDECLARATION_START error { force_error (Error_xml_declaration) }
	;

encoding_decl: XMLDECLARATION_ENCODING
	;

-- 4.7 Notation declaration

notation_decl: DOCTYPE_NOTATION req_space doctype_name req_space external_id maybe_space DOCTYPE_END
	| DOCTYPE_NOTATION req_space doctype_name req_space public_id maybe_space DOCTYPE_END
	| DOCTYPE_NOTATION error { force_error (Error_doctype_notation) }
	;

public_id: DOCTYPE_PUBLIC
		{ $$ := $1 }
	;

%%
feature {NONE} -- creation

	make is
			-- Initialise.
		do
			make_scanner
			make_parser

			-- parser state
			stand_alone := False
			in_external_dtd := False
			
			-- entities
			!! entities.make_default
			!! pe_entities.make_default
		end

	-- reset?

feature {NONE} -- State

	stand_alone: BOOLEAN
			-- From XML declaration.

	entities: DS_HASH_TABLE [NXML_ENTITY_DEF,STRING]
			-- Defined entites
	pe_entities: DS_HASH_TABLE [NXML_ENTITY_DEF,STRING]
			-- Defined PE entities
	
	in_external_dtd: BOOLEAN
			-- Is the scanner in an external DTD?
		
feature {NONE} -- Factory

	new_namespace_name: NXML_PARSER_NAME is
			-- Create namespace_name.
		do
			!! Result.make
		ensure
			new_object: Result /= Void
			empty: Result.is_empty
		end
		
	new_name_set: DS_HASH_SET [NXML_PARSER_NAME] is
			-- Name set for checking.
		do
			!! Result.make_equal (0)
		end

feature {NONE} -- Entitties

	new_literal_entity (a_value: STRING): NXML_ENTITY_DEF is
			-- New literal entity definition.
		require
			not_void: a_value /= Void
		do
			!! Result.make_literal (a_value)
		end
		
	new_external_entity (a_value: STRING): NXML_ENTITY_DEF is
			-- New external entity definition.
		require
			not_void: a_value /= Void
		do
			!! Result.make_external (a_value)
		end
		
feature {NXML_PARSER} -- Entities

	when_entity_declared (a_name: STRING; a_def: NXML_ENTITY_DEF) is
			-- Entity has been declared.
		require
			name_not_void: a_name /= Void
		do
		 	debug ("nxml")
				io.put_string ("Entity declared: ")
				io.put_string (a_name)
				io.put_string (" value: ")
				io.put_string (a_def.value)
				io.put_new_line
			end
			
			-- 4.2: when multiple declaration first one is binding
			if a_def /= Void then
				if not entities.has (a_name) then
					entities.force (a_def, a_name)
				end
				check has: entities.has (a_name) end
			end
		end
		
	when_pe_entity_declared (a_name: STRING; in_def: NXML_ENTITY_DEF) is
			-- PE entity has been declared.
		require
			name_not_void: a_name /= Void
		local
			a_def: NXML_PE_ENTITY_DEF
		do
		 	debug ("nxml")
				io.put_string ("PE entity declared: ")
				io.put_string (a_name)
				io.put_string (" value: ")
				io.put_string (in_def.value)
				io.put_new_line
			end

			-- convert to PE
			!! a_def.make_def (in_def)
			
			-- 4.2: when multiple declaration take first one
			if a_def /= Void then
				if not pe_entities.has (a_name) then
					pe_entities.force (a_def, a_name)
				end
				check has: pe_entities.has (a_name) end
			end
		end
		
	entity_referenced_in_entity_value (a_name: STRING): STRING is
			-- Named parameter entity has been referenced in entity value.
		require
			not_void: a_name /= Void
		do
			if in_external_dtd then
				if pe_entities.has (a_name) then
					Result := defined_entity_referenced (pe_entities.item (a_name))
				else
					-- 4.4.4 Forbidden
					force_error (Error_external_reference_in_quoted_value)
				end
			else
				force_error (Error_doctype_peref_only_in_dtd)
			end
		end

	defined_entity_referenced (a_def: NXML_ENTITY_DEF): STRING is
			-- Resolved defined entity in quoted value.
		require
			not_void: a_def /= Void
		do
			if a_def.is_literal then
				-- 4.4.5 Included in literal
				Result := a_def.value
			else
				Result := external_entity_to_string (a_def.value)
			end
		end
		
	external_entity_to_string (a_sys: STRING): STRING is
			-- External entity to string.
		require
			not_void: a_sys /= Void
		local
			a_file: like INPUT_STREAM_TYPE
			an_error: STRING
		do
			!! Result.make (0)
			a_file := INPUT_STREAM_.make_file_open_read (a_sys)
			if INPUT_STREAM_.is_open_read (a_file) then
				-- inefficient?
				from until
					INPUT_STREAM_.end_of_input (a_file)
				loop
					Result.append_string (INPUT_STREAM_.read_string (a_file,100))
				end
			else
				!! an_error.make (0)
				an_error.copy (Error_cannot_read_file)
				an_error.append_string (": ")
				an_error.append_string (a_sys)
				force_error (an_error)
			end
		end
		
feature {NONE} -- DTD

	when_external_dtd (a_system: STRING) is
			-- Load external DTD from system name.
		require
			system_ok: a_system /= Void
		do
			debug ("nxml") io.put_string ("[when_external_dtd]") end
			
			-- push old scanner
			scanners.force (scanner)
			!NXML_SCANNER_DTD! scanner.make_scanner
			scanner.set_input_file (a_system)
			if not scanner.last_input_file_opened then
				when_error (Error_cannot_open_external_dtd)
			end
		end

feature -- Parser

	parse_file (a_name: STRING) is
			-- Parse file.
		require
			not_void: a_name /= Void
		do
			scanner.set_input_file (a_name)
			if scanner.last_input_file_opened then
				parse
			else
				force_error (Error_cannot_read_file)
			end
		end
		
	parse_string (a_string: STRING) is
			-- Parse string.
		require
			not_void: a_string /= Void
		do
			scanner.set_input_buffer (scanner.new_string_buffer (a_string))
			parse
		end
		
feature -- Errors
	
	force_error (a_message: STRING) is
			-- Report error message.
		do
			when_error (a_message)
			abort
		end

feature {NONE} -- Scanner implementation

	make_scanner is
			-- Initialise main scanner and saved
			-- scannners stack.
		do
			!NXML_SCANNER! scanner.make_scanner
			!! scanners.make
		end
		
	scanner: NXML_SCANNER
			-- Scanner for PE entity.

	scanners: DS_LINKED_STACK[NXML_SCANNER]
			-- Scanner stack for entities etc.

	last_token: INTEGER
			-- Last token read by read_token.
	last_value: ANY
			-- Last value read by read_token.
	
	read_token is
			-- Read token from scanner.
		do
			-- read token from scanner
			scanner.read_token
			last_token := scanner.last_token
			last_value := scanner.last_value
			
			debug ("nxml")
				io.put_string (token_name (last_token))
				io.put_string ("/")
				if last_value /= Void then
					io.put_string (last_value.out)
					io.put_string ("!")
				end
			end
			
			-- unwind scanner stack if end of current one
			if scanner.end_of_file and not scanners.is_empty then
				scanner := scanners.item
				scanners.remove
				read_token
			end
			
			-- if this is a PE entity reference, temporarily
			-- switch scanner. Token is left for validation.
			if last_token = DOCTYPE_PEREFERENCE then
				process_pe_entity (last_value.out)
			-- same if content entity reference
			elseif last_token = CONTENT_ENTITY  then
				process_entity (last_value.out)
			elseif last_token = ATTRIBUTE_ENTITY then
				process_attribute_entity (last_value.out)
			end
		end

feature {NONE} -- Scanner entity processing

	process_pe_entity (a_name: STRING) is
			-- Pushn PE entity scanner on stack.
		require
			not_void: a_name /= Void
		do
			if pe_entities.has (a_name) then
				process_entity_scanner (pe_entities.item (a_name))
			else
				force_error (Error_doctype_undefined_pe_entity)
			end
		end
		
	process_entity (a_name: STRING) is
			-- Push entity scanner on stack.
		require
			not_void: a_name /= Void
		do
			if entities.has (a_name) then
				process_entity_scanner (entities.item (a_name))
			else
				-- 4.4.4 Forbidden
				force_error (Error_entity_undefined)
			end
		end
	
	process_attribute_entity (a_name: STRING) is
			-- Push entity scanner on stack.
		require
			not_void: a_name /= Void
		do
			if entities.has (a_name) then
				if entities.item (a_name).is_literal then
					process_entity_scanner (entities.item (a_name))
				else
					force_error (Error_entity_literal_in_attribute)
				end
			else
				-- 4.4.4 Forbidden
				force_error (Error_entity_undefined)
			end
		end
		
	process_entity_scanner (a_def: NXML_ENTITY_DEF) is
			-- Save previous scannner, create new scanner, and
			-- reset it with previous scanner state. 
		require
			not_void: a_def /= Void
		do				
			a_def.apply_input_buffer
				
			if a_def.has_error then
				force_error (a_def.last_error)
			else
				-- push scanner
				a_def.set_start_condition (scanner.start_condition)
				scanners.force (scanner)
				scanner := a_def
			end
		end

feature -- Error diagnostic

	line: INTEGER is
			-- Current line.
		do
			Result := scanner.line
		end
		
	column: INTEGER is
			-- Current column.
		do
			Result := scanner.column
		end
		
	filename: STRING is
			-- Current file.
		do
			Result := scanner.filename
		end
		
	error_header: STRING is
			-- Header for error message.
			-- <filename>:<line>:<column>:
		do
			!! Result.make (0)
			Result.append_string (filename)
			Result.append_string (":")
			Result.append_string (line.out)
			Result.append_string (":")
			Result.append_string (column.out)
			Result.append_string (":")
		ensure
			not_void: Result /= Void
		end

feature {NONE} -- String

	concatenate (a, b: STRING): STRING is
			-- Result := a + b
		require
			not_void: a /= Void and b /= Void
		do
			!! Result.make (0)
			Result.append_string (a)
			Result.append_string (b)
		end

	empty_string: STRING is
			-- Shared empty string. 
		once
			!!Result.make (0)
		end

invariant
	scanner_not_void: scanner /= Void
	scanners_not_void: scanners /= Void
		
end
